from typing import Union
from .batch import Batch, Batchable, _batch_repr


class DFRow(Batchable):
    def __init__(self, **kwargs):
        for key, val in kwargs.items():
            self.__dict__[key] = val

    def __repr__(self):
        args = []
        for key, val in self.asdict().items():
            val_str = _batch_repr(val)
            args.append(f"{key}={val_str}")
        return f"DFRow({', '.join(args)})"

    def __getitem__(self, key):
        return self.__dict__[key]

    def __setitem__(self, key, val):
        self.__dict__[key] = val

    def keys(self):
        return self.__dict__.keys()

    def values(self):
        return self.__dict__.values()

    def items(self):
        return self.__dict__.items()

def merge(items: Union[DFRow, Batch]):
    template = items[0]
    attribs = {}
    for item in items:
        attribs.update(item.asdict())
    if isinstance(template, Batch):
        return Batch(DFRow, **attribs)
    else:
        return DFRow(**attribs)


from functools import partial
from torch.utils._pytree import tree_map
from typing import Any, Generic, Optional, Type, TypeVar, Union, List, Tuple
from dataclassy import dataclass
import torch
import torch.utils.data
from torch.utils.data.dataloader import default_collate
from .categorical_tensor import CategoricalTensor, NoStackCatTensor
from .meta_utils import recursive_map


@dataclass
class Batchable:
    """ Base class for all objects we want to batchify
    
    This method can also define static methods collate_{attribute}
    and index_{attribute} """
    
    @staticmethod
    def get_batch_type():
        """ override this method if you want a custom batch type
        for your batchable class """
        return Batch

    def asdict(self):
        """ Override this method to define what attributes you want your
        Batch to define """
        return self.__dict__

T = TypeVar('T')

class BatchBase(Generic[T]):

    def item_type(self) -> Type[Batchable]:
        """ Returns the type of each item in the batch """
        raise NotImplementedError()

    def cuda(self):
        return self.to("cuda")

    def cpu(self):
        return self.to("cpu")

def _batch_repr(val):
    if isinstance(val, torch.Tensor):
        val_str = f"Tensor(shape={val.shape}, dtype={val.dtype})"
    else:
        val_str = repr(val)
    return val_str

class Batch(BatchBase[T]):

    _internal_attribs = [ "_batch_len", "_batch_type", "__dict__" ]
    _batch_len: int
    _batch_type: Type[Batchable]

    def __init__(self,
                items: Optional[Union[List[T], Type[T]]],
                **kwargs):
        if isinstance(items, type):
            self._init_from_type(items, **kwargs)
        else:
            self._init_from_list(items)

    def _init_from_list(self, items: List[T]):

        assert len(items) > 0
        template = items[0]
        template_type = template.__class__ # type(template)
        # todo: very hacky
        if isinstance(template, BatchViewBase):
            template_type = template.get_type()

        self._batch_len = len(items)
        self._batch_type = template_type

        if isinstance(template, BatchBase):
            raise ValueError("Batchifying batches is not supported at the moment")

        attribs = { key: [] for key in template.asdict().keys() }
        for item in items:
            for key, attrib in item.asdict().items():
                attribs[key].append(attrib)

        for key, attrib_list in attribs.items():

            if key in Batch._internal_attribs:
                raise ValueError(f"{key} is used internally by Batch, so it shouldn't be a member of {template_type}")

            collate_method = "collate_" + key
            if hasattr(template_type, collate_method):
                collated = getattr(template_type, collate_method)(attrib_list)
            else:
                collated = collate(attrib_list)

            self.__dict__[key] = collated
        

    def _init_from_type(self, batch_type: Type, **kwargs):
        self._batch_type = batch_type
        if len(kwargs) == 0:
            raise ValueError("Can't determine batch size of empty batch")
        self._batch_len = len(next(iter(kwargs.values())))
        for key, val in kwargs.items():
            self.__dict__[key] = val

    def __len__(self) -> int:
        return self._batch_len

    def __getitem__(self, index) -> Union["BatchView[T]", Any]:
        if isinstance(index, str):
            return self.__dict__[index]
        if isinstance(index, int) and index >= len(self):
            raise IndexError()
        if isinstance(index, int):
            return BatchView(self, index)
        return collate([self[i] for i in range(len(self))[index]])

    def item_type(self) -> Type[T]:
        """ Returns the type of each item in the batch """
        return self._batch_type

    def attribute_names(self) -> List[str]:
        """ Returns the names of all the batched attributes """
        return [ key for key in self.__dict__.keys() if key not in Batch._internal_attribs ]

    def asdict(self):
        """ Convert to dict """
        return { key: val for key, val in self.__dict__.items() if key not in Batch._internal_attribs }

    def to(self, device):
        to_dict = { key: recursive_map(lambda x: x.to(device) if hasattr(x, 'to') else x, val) for key, val in self.asdict().items() }
        return Batch(self._batch_type, **to_dict)

    def __repr__(self):
        args = []
        for key, val in self.asdict().items():
            val_str = _batch_repr(val)
            args.append(f"{key}={val_str}")
        return f"Batch[{self._batch_type.__name__}]({', '.join(args)})"

    def __getattribute__(self, name: str) -> Any:
        if name in Batch._internal_attribs or name in self.__dict__:
            return object.__getattribute__(self, name)
        batch_method_name = "batch_" + name
        if hasattr(self, "_batch_type") and hasattr(self._batch_type, batch_method_name):
            batch_method = getattr(self._batch_type, batch_method_name)
            if callable(batch_method):
                return partial(batch_method, self)
        return object.__getattribute__(self, name)

class BatchViewBase(Generic[T], Batchable):
    pass

def _get_methods_for_type(type_):
    return {func: getattr(type_, func) for func in dir(type_) if callable(getattr(type_, func)) and not func.startswith("__")}

class BatchView(BatchViewBase[T]):
    """ View of an item in a batch. Should act like said item in most
    circumstances. We use views instead of creating actual items because,
    for many use cases, lazily indexing batches is much faster """

    _internal_attribs = [ "_batch", "_index", "_get_methods" ]
    _batch: Batch[T]
    _index: Union[int, slice] # either int or slice

    def __init__(self, batch: Batch, index: Union[int, slice]):
        self._batch = batch
        self._index = index
        # self.__class__ = self._batch._batch_type

    def _get_methods(self):
        type_ = self._batch._batch_type
        return { key: val for key, val in _get_methods_for_type(type_).items() if key not in _get_methods_for_type(BatchView) }

    def __getattribute__(self, name: str) -> Any:
        if name == "__dict__" or name in BatchView._internal_attribs or name in _get_methods_for_type(BatchView):
            return object.__getattribute__(self, name)
        if name in self.__dict__:
            return object.__getattribute__(self, name)
        if name in self._batch.__dict__ and not name in Batch._internal_attribs:
            attrib = getattr(self._batch, name)

            item_type = self._batch.item_type()
            index_method = "index_" + name
            if hasattr(item_type, index_method):
                ret = getattr(item_type, index_method)(attrib, self._index)
                self.__dict__[name] = ret
                return ret

            if isinstance(attrib, tuple):
                ret = tuple([ item[self._index] for item in attrib ])
                self.__dict__[name] = ret
                return ret
            elif isinstance(attrib, dict):
                ret = { key: val[self._index] for key, val in attrib.items() }
                self.__dict__[name] = ret
                return ret

            ret = attrib[self._index]
            self.__dict__[name] = ret
            return ret

        if isinstance(self._index, int):
            # assume we are acting like a T
            methods = self._get_methods()
            if name in methods:
                return partial(methods[name], self)
        else:
            # assume we are acting like a Batch[T]
            # todo: make this whole "what am I " process more robust
            batch_method_name = "batch_" + name
            if hasattr(self, "_batch") and hasattr(self.get_type(), batch_method_name):
                batch_method = getattr(self.get_type(), batch_method_name)
                if callable(batch_method):
                    return partial(batch_method, self)
            pass
        return object.__getattribute__(self, name)

    def __getitem__(self, index):
        if isinstance(self._index, int):
            raise ValueError("Can only index into a BatchView if the batchview is view of a slice of a batch")
        new_idx = range(len(self._batch))[index]
        if isinstance(new_idx, range):
            new_idx = slice(new_idx.start, new_idx.stop, new_idx.step)
            raise NotImplementedError # need to test this
        else:
            assert isinstance(new_idx, int)
        return BatchView(self._batch, new_idx)

    def asdict(self):
        return { key: getattr(self, key) for key in self._batch.attribute_names() }

    def __repr__(self):
        args = []
        for key, val in self.asdict().items():
            val_str = _batch_repr(val)
            args.append(f"{key}={val_str}")
        return f"BatchView[{self.get_type().__name__}]({', '.join(args)})"

    def get_type(self):
        return self._batch._batch_type

class NoStackTensor(torch.Tensor):
    """ This is used when you want to collate tensors into a list instead
    of stack them. E.g. when you have different shapes"""

    @staticmethod
    def __new__(cls, tensor: torch.Tensor):
        return torch.Tensor._make_subclass(cls, tensor.to('meta'))

    def __init__(self, tensor):
        self.tensor = tensor

    @classmethod
    def __torch_function__(cls, func, types, args=(), kwargs={}):
        # return torch.Tensor.__torch_dispatch__(func, types, args, kwargs)
        def unwrap(x):
            return x.tensor if isinstance(x, NoStackTensor) else x
        def wrap(x):
            return NoStackTensor(x) if isinstance(x, torch.Tensor) else x
        args = tree_map(unwrap, args)
        kwargs = tree_map(unwrap, kwargs)
        out = func(*args, **kwargs)
        return tree_map(wrap, out)

def collate(batch: Any) -> Any:
    """ turn a list of items into a batch of items. Replacement
    for pytorch's default collate. This is what we use in the
    custom DataLoader class """
    # performance optimization -- if we've already batched something, no
    # need to do it again
    if isinstance(batch, Batch):
        return batch

    example = batch[0]
    if isinstance(example, Batchable):
        return type(example).get_batch_type()(batch)
    elif isinstance(example, tuple) or isinstance(example, list):
        ret = []
        for i, item in enumerate(example):
            all_items = [ b[i] for b in batch]
            ret.append(collate(all_items))
        return type(example)(ret)
    elif isinstance(example, dict):
        ret = {}
        for key in example.keys():
            to_collate = []
            for item in batch:
                to_collate.append(item[key])
            ret[key] = collate(to_collate)
        return ret
    elif isinstance(example, NoStackTensor) or isinstance(example, NoStackCatTensor):
        return batch
    elif isinstance(example, CategoricalTensor):
        return torch.stack(batch)
    else:
        try:
            return default_collate(batch)
        except TypeError:
            return batch

class DataLoader(torch.utils.data.DataLoader):
    """ Dataloader that correctly batchifies Batchable data """
    
    def __init__(self, dataset, batch_size=1, shuffle=False, **kwargs):
        super(DataLoader, self).__init__(
            dataset, batch_size, shuffle,
            collate_fn=collate, **kwargs
        )


import torch
import torch.nn as nn

from .categorical_tensor import CategoricalTensor

class Module(nn.Module):

    def __init__(self):
        super().__init__()
        self._initialized = False
        self._started_forward = False
        self._submodule_index = 0
        self._submodules = nn.ModuleList()
        self._param_index = 0
        self._params = nn.ParameterList()
        self._checkpoints = []
        self._checkpoint_index = 0

    def start_forward(self):
        # todo: these boolean values getting unweildly -- cut down!
        if self._submodule_index != 0:
            self._initialized = True
        self._submodule_index = 0
        self._param_index = 0
        self._started_forward = True
        self._checkpoint_index = 0

    def make(self, cls, *args, **kwargs):
        if not self._started_forward:
            raise RuntimeError("You must call start_forward before you call make")
        if not self._initialized:
            self._submodules.append(cls(*args, **kwargs))
        submod = self._submodules[self._submodule_index]
        self._submodule_index += 1
        return submod

    def make_param(self, cls, *args, **kwargs):
        if not self._started_forward:
            raise RuntimeError("You must call start_forward before you call make_param")
        if not self._initialized:
            self._params.append(cls(*args, **kwargs))
        param = self._params[self._param_index]
        self._param_index+= 1
        return param

    def checkpoint(self):
        if self._initialized:
            self._submodule_index, self._param_index = self._checkpoints[self._checkpoint_index]
            self._checkpoint_index += 1
        else:
            self._checkpoints.append((self._submodule_index, self._param_index))

    def is_initialized(self):
        return self._initialized or self._submodule_index > 0 or self._param_index > 0

    def parameters(self, recurse: bool = True):
        assert self.is_initialized(), "Terrace Module needs to be run on data before parameters method can be called"
        return super().parameters(recurse)

class WrapperModule(Module):

    def __init__(self, *args, **kwargs):
        super().__init__()
        self.args = args
        self.kwargs = kwargs

class LazyLinear(WrapperModule):
    """ torch >=1.13 has this, but wanted to try my hand at implementing it myself
    (and ensuring that older torch versions work) """

    def forward(self, x):
        self.start_forward()
        in_feats = x.shape[-1]
        return self.make(nn.Linear, in_feats, *self.args, **self.kwargs)(x)


class LazyEmbedding(Module):

    def __init__(self, embedding_dims):
        super().__init__()
        self.embedding_dims = embedding_dims

    def forward(self, x: CategoricalTensor):
        self.start_forward()
        embedding_dims = [self.embedding_dims]*len(x.num_classes) if isinstance(self.embedding_dims, int) else self.embedding_dims
        ret = []
        for idx in range(x.shape[-1]):
            max_val = x.num_classes[idx]
            ret.append(self.make(nn.Embedding, max_val, embedding_dims[idx])(x.tensor[..., idx]))
        return torch.cat(ret, -1)

class LazyMultiheadAttention(WrapperModule):

    def forward(self, q, k, v):
        self.start_forward()
        embed_dim = q.shape[-1]
        self.kwargs["kdim"] = k.shape[-1]
        self.kwargs["vdim"] = v.shape[-1]
        return self.make(nn.MultiheadAttention, embed_dim, *self.args, **self.kwargs)(q, k, v)

class LazyLayerNorm(WrapperModule):

    def forward(self, x):
        self.start_forward()
        return self.make(nn.LayerNorm, x.shape[1:], *self.args, **self.kwargs)(x)



# utility functions for python metaprogramming
from typing import Any, Type, Callable
from dataclasses import dataclass
from copy import deepcopy

def default_init(cls):
    """ makes a default init function for a class """
    return dataclass(eq=False,repr=False)(cls)

class classclass():
    """ Allows a class within a class to access subclasses
    of the owner class """
    def __init__(self, cls):
        self.cls = cls
    def __get__(self, instance, cls):
        new_cls = deepcopy(self.cls)
        new_cls.Owner = cls
        return new_cls

def get_type_name(typ):
    """ Returns the name of a type even if it's a generic alias
    (because python is weird and something like List[int] won't have
    a __name__ but List will) """
    if hasattr(typ, "__name__"):
        return typ.__name__
    elif hasattr(typ, "__origin__"):
        return typ.__origin__.__name__ + "[" + ", ".join(map(get_type_name, typ.__args__)) + "]"


def get_any_arg(*args, **kwargs):
    """ Returns the first argument given, regardless of kwarg or not """
    if len(args) > 0:
        return args[0]
    if len(kwargs) > 0:
        return next(iter(kwargs.values))
    raise AssertionError

def contains_type(item: Any, T: Type):
    """ Returns true if item is an instance of T or
    is a container that contains T """
    if isinstance(item, T):
        return True
    elif isinstance(item, list) or isinstance(item, tuple):
        for i2 in item:
            if contains_type(i2, T):
                return True
    elif isinstance(item, dict):
        for i2 in item.values():
            if contains_type(i2, T):
                return True
    return False

def recursive_map(func: Callable, item: Any):
    """ recurisvely applies func to all non-container
    items in the container """
    
    if isinstance(item, list) or isinstance(item, tuple):
        return type(item)([ recursive_map(func, i2) for i2 in item ])
    elif isinstance(item, dict):
        return { key: recursive_map(func, val) for key, val in item.items() }
    else:
        return func(item)

def recursive_zip(c1, c2):
    """ c1 and c2 are arbitrary containers. They must have the same
    "same" e.g. lists have the same len, dicts have the same keys etc.
    this yeilds the zipped leaves of c1 and c2"""
    
    if isinstance(c1, list) or isinstance(c1, tuple):
        for i1, i2 in zip(c1, c2):
            for tup in recursive_zip(i1, i2): yield tup
    elif isinstance(c1, dict):
        for key in c1:
            for tup in recursive_zip(c1[key], c2[key]): yield tup
    else:
        yield c1, c2


from .module import *
from .categorical_tensor import CategoricalTensor, NoStackCatTensor
from .graph import Graph, GraphBatch
from .batch import Batchable, BatchBase, Batch, collate, DataLoader, NoStackTensor
from .dataframe import DFRow


from copy import deepcopy
from typing import Tuple, Union
import functools
import torch

from torch.utils._pytree import tree_map

# https://pytorch.org/docs/stable/notes/extending.html

# https://colab.research.google.com/drive/1MLeSCMrc6a5Yf_6uAh0bH-IPf5c_BTQf#scrollTo=dM0sl2X6epBX

NumClassesType = Union[int, Tuple[int, ...]]

HANDLED_FUNCTIONS = {}
# if num_classes is a tuple, it decribes the number of classes along the last
# dimension of the tensor
class CategoricalTensor(torch.Tensor):
    @staticmethod
    def __new__(cls, tensor: torch.Tensor, num_classes: NumClassesType):
        return torch.Tensor._make_subclass(cls, tensor.to('meta'))

    def __init__(self, tensor, num_classes: NumClassesType):
        assert tensor.dtype == torch.long
        if isinstance(num_classes, int):
            num_classes = tuple([num_classes]*tensor.shape[-1])
        assert len(num_classes) == tensor.shape[-1]
        self.tensor = tensor
        self.num_classes = num_classes

    def __repr__(self):
        return f"{self.__class__.__name__}(tensor={self.tensor}, num_classes={self.num_classes})"

    def __getstate__(self):
        return (self.tensor, self.num_classes)

    def __setstate__(self, state):
        self.tensor, self.num_classes = state

    # Only difference with above is to add this function.
    @classmethod
    def __torch_function__(cls, func, types, args=(), kwargs=None):
        if kwargs is None:
            kwargs = {}
        if func in HANDLED_FUNCTIONS:
            return HANDLED_FUNCTIONS[func](*args, **kwargs)
        # return func(*args, **kwargs)
        raise NotImplementedError(func)

class NoStackCatTensor(CategoricalTensor):
    pass


def implements(torch_function):
    """Register a torch function override for ScalarTensor"""
    @functools.wraps(torch_function)
    def decorator(func):
        HANDLED_FUNCTIONS[torch_function] = func
        return func
    return decorator

@implements(torch.Tensor.__deepcopy__)
def ct_deepcopy(ct, memo):
    return CategoricalTensor(deepcopy(ct.tensor, memo), deepcopy(ct.num_classes, memo))

@implements(torch.split)
def split(ct, *args, **kwargs):
    """ No error checking for now"""
    ret = torch.split(ct.tensor, *args, **kwargs)
    return [ CategoricalTensor(t, ct.num_classes) for t in ret ]

@implements(torch.Tensor.__len__)
def cat_tensor_len(ct):
    return len(ct.tensor)

@implements(torch.Tensor.cuda)
def cuda(ct):
    return CategoricalTensor(ct.tensor.cuda(), num_classes=ct.num_classes)

@implements(torch.Tensor.cpu)
def cpu(ct):
    return CategoricalTensor(ct.tensor.cpu(), num_classes=ct.num_classes)

@implements(torch.Tensor.to)
def to(ct, device):
    return CategoricalTensor(ct.tensor.to(device), num_classes=ct.num_classes)

def construct_cat_tensor(*args):
    num_classes, (tensor_func, tensor_args) = args
    tensor = tensor_func(*tensor_args)
    return CategoricalTensor(tensor, num_classes=num_classes)

@implements(torch.Tensor.__reduce_ex__)
def cat_reduce_ex(self, proto):
    return (construct_cat_tensor, (self.num_classes, self.tensor.__reduce_ex__(proto)))

@implements(torch.cat)
def cat(inputs: Tuple[CategoricalTensor, ...], dim: int = -1) -> CategoricalTensor:
    if dim < 0:
        dim = len(inputs[0].shape) - dim
    if dim == len(inputs[0].shape):
        num_classes = []
        for t in inputs:
            num_classes += t.num_classes
    else:
        num_classes = inputs[0].num_classes
        for inp in inputs:
            assert inp.num_classes == num_classes
    tensor = torch.cat([t.tensor for t in inputs], dim)
    return CategoricalTensor(tensor, num_classes)

@implements(torch.stack)
def stack(inputs: Tuple[CategoricalTensor, ...]):
    assert len(inputs) > 0
    num_classes = inputs[0].num_classes
    for t in inputs:
        assert t.num_classes == num_classes
    return CategoricalTensor(torch.stack([t.tensor for t in inputs]), num_classes)

@implements(torch.Tensor.__getitem__)
def getitem(t: CategoricalTensor, idx: int):
    num_classes = t.num_classes
    if isinstance(idx, tuple):
        if len(idx) == len(t.shape):
            num_classes = t.num_classes[idx[-1]]
    else:
        if len(t.shape) == 1:
            num_classes = t.num_classes[idx]
    if isinstance(num_classes, int):
        num_classes = [ num_classes ]
    return CategoricalTensor(t.tensor[idx], num_classes)

@implements(torch.Tensor.shape.__get__)
def get_shape(t: CategoricalTensor):
    return t.tensor.shape

@implements(torch.Tensor.device.__get__)
def get_device(t: CategoricalTensor):
    return t.tensor.device

@implements(torch.Tensor.dtype.__get__)
def get_dtype(t: CategoricalTensor):
    return t.tensor.dtype


from typing import Dict, List, Sequence, Set, Tuple, Optional, Type, TypeVar, Generic, Any, Union
from dataclassy import dataclass
import dgl
import torch

from .batch import BatchView, BatchViewBase, Batchable, Batch, BatchBase, collate

@dataclass
class TypeTree:
    type: Type
    subtypes: Dict[str, "TypeTree"]

def flatten_batch(b: Batch) -> Tuple[Dict[str, torch.Tensor], TypeTree]:
    """ Flattens batch to a single dict from keys to tensors. Also returns
    all the batch types. This is neccessary because dgl stores ndata and edata
    as such dicts. """

    type_tree = TypeTree(b.item_type(), {})
    ret = {}

    for key, val in b.asdict().items():
        if isinstance(val, torch.Tensor):
            ret[key] = val
            type_tree.subtypes[key] = TypeTree(torch.Tensor, {})
        elif isinstance(val, Batch):
            sub_dict, subtype_tree = flatten_batch(val)
            type_tree.subtypes[key] = subtype_tree
            for key2, val2 in sub_dict.items():
                ret[f"{key}/{key2}"] = val2
        else:
            raise NotImplementedError(f"flatten_batch is only currently implemented for torch.Tensor and Batches, but found attribute {key} with type {type(val)}")

    return ret, type_tree

def unflatten_dict(flat):
    """ Helper function for unflatten_batch. Returns an unflattened dict
    for a dict with keys like 'key/subkey' """
    ret = {}
    for key, val in flat.items():
        container = ret
        subkeys = key.split("/")
        for subkey in subkeys[:-1]:
            if subkey not in container:
                container[subkey] = {}
            container = container[subkey]
        if val == 'None':
            val = None
        container[subkeys[-1]] = val
    return ret

def make_batch_from_unflat_dict(unflat_dict, type_tree):
    """ Helper function for unflatten_batch. Makes a batch from
    a dict that has been unflattened """
    batch_type = type_tree.type
    kwargs = {}
    for key, val in unflat_dict.items():
        if isinstance(val, dict):
            arg = make_batch_from_unflat_dict(val, type_tree.subtypes[key])
        elif isinstance(val, torch.Tensor):
            arg = val
        else:
            raise AssertionError(f"Expected only dicts and tensors, but got {key}={val}")
        kwargs[key] = arg
    return Batch(batch_type, **kwargs)

def unflatten_batch(flat_dict: Dict[str, torch.Tensor], type_tree: TypeTree) -> Batch:
    """ Reverses flatten_batch; takes a flattened dict and a type tree and
    returns a proper Batch """
    unflat_dict = unflatten_dict(flat_dict)
    return make_batch_from_unflat_dict(unflat_dict, type_tree)

N = TypeVar('N', bound=Batchable)
E = TypeVar('E', bound=Optional[Batchable])

@dataclass
class GraphBase(Generic[N, E]):

    _dgl_graph: Union[dgl.graph, dgl.batch]
    _node_type_tree: TypeTree
    _edge_type_tree: Optional[TypeTree]

    @property
    def ndata(self) -> Batch[N]:
        return unflatten_batch(self._dgl_graph.ndata, self._node_type_tree)

    @property
    def edata(self) -> Batch[E]:
        return unflatten_batch(self._dgl_graph.edata, self._edge_type_tree)

    @property
    def edges(self) -> List[Tuple[int, int]]:
        ret = []
        for src, dst in zip(*self._dgl_graph.edges()):
            ret.append((int(src), int(dst)))
        return ret

    def dgl(self) -> Union[dgl.graph, dgl.batch]:
        return self._dgl_graph

    def to(self, device):
        cls_ = type(self)
        ret = cls_.__new__(cls_)
        for key, val in self.__dict__.items():
            if key == "_dgl_graph":
                val = val.to(device)
            ret.__dict__[key] = val
        return ret

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(ndata={self.ndata}, edata={self.edata})"


class Graph(GraphBase[N, E], Batchable):
    """ Wrapper around dgl graph allowing easier access to data """

    @staticmethod
    def get_batch_type():
        return GraphBatch

    def __init__(self, nodes: Sequence[N],
                 edges: Sequence[Tuple[int, int]],
                 edata: Optional[List[E]] = None,
                 directed: bool = False,
                 device = 'cpu'):
        """ If directed is false, both permutations of the edges
        will be added. """

        src_list = []
        dst_list = []
        new_edata = None if edata is None else []
        for i, (n1, n2) in enumerate(edges):
            src_list.append(n1)
            dst_list.append(n2)
            # if new_edata is not None:
            #     new_edata.append(edata[i])
            if not directed:
                src_list.append(n2)
                dst_list.append(n1)
                if new_edata is not None:
                    new_edata.append(edata[i])
                    new_edata.append(edata[i])
        
        if directed and edata is not None:
            assert new_edata == []
            new_edata = edata
        
        dgl_graph = dgl.graph((torch.tensor(src_list), torch.tensor(dst_list)), num_nodes=len(nodes), idtype=torch.int32, device=device)

        node_batch = collate(nodes)
        node_flat_dict, node_type_tree = flatten_batch(node_batch)
        for key, val in node_flat_dict.items():
            dgl_graph.ndata[key] = val

        if new_edata is not None:
            edge_batch = collate(new_edata)
            edge_flat_dict, edge_type_tree = flatten_batch(edge_batch)
            for key, val in edge_flat_dict.items():
                dgl_graph.edata[key] = val
        else:
            edge_type_tree = None

        super().__init__(dgl_graph, node_type_tree, edge_type_tree)

G = TypeVar('G', bound=Graph)
class GraphBatch(BatchBase[G], GraphBase):

    _graph_type: Type[Graph]
    node_slices: List[slice]
    edge_slices: List[slice]
    
    def __init__(self, items: List[Graph[N, E]]):
        dgl_batch = dgl.batch([ item.dgl() for item in items ])
        first = items[0]
        super().__init__(dgl_batch, first._node_type_tree, first._edge_type_tree)
        self._graph_type = type(first)

        # node and edge slices determine which nodes/edges belong to which
        # graph in the batch
        self.node_slices = []
        self.edge_slices = []
        tot_n = 0
        tot_e = 0
        for n, e in zip(self.dgl().batch_num_nodes(), self.dgl().batch_num_edges()):
            n = int(n)
            e = int(e)
            self.node_slices.append(slice(tot_n,tot_n+n))
            self.edge_slices.append(slice(tot_e,tot_e+e))
            tot_n += n
            tot_e += e

    def __len__(self) -> int:
        return self.dgl().batch_size

    def __getitem__(self, index: int) -> Graph[N, E]:
        if isinstance(index, int):
            if index >= len(self):
                raise IndexError()
        else:
            raise NotImplementedError()
        return GraphBatchView(self, index)

    def item_type(self):
        return self._graph_type

class GraphBatchView(BatchViewBase[G]):

    _internal_attribs = [ "_batch", "_index" ]
    _batch: Batch[G]
    _index: Union[int, slice] # either int or slice
    _node_type_tree: TypeTree
    _edge_type_tree: TypeTree

    def __init__(self, batch: Batch, index: Union[int, slice]):
        self._batch = batch
        self._index = index
        self._node_type_tree = batch._node_type_tree
        self._edge_type_tree = batch._edge_type_tree

    @property
    def ndata(self) -> Batch[N]:
        idxs = self._batch.node_slices[self._index]
        return self._batch.ndata[idxs]

    @property
    def edata(self) -> Batch[E]:
        idxs = self._batch.edge_slices[self._index]
        return self._batch.edata[idxs]

    @property
    def edges(self) -> List[Tuple[int, int]]:
        node_idxs = self._batch.node_slices[self._index]
        n0 = node_idxs.start
        node_idxs_expl = range(len(self._batch.ndata))[node_idxs]
        ret = []
        for n1, n2 in self._batch.edges:
            if n1 in node_idxs_expl:
                assert n2 in node_idxs_expl
                ret.append((n1 - n0, n2 - n0))
        return ret

    def dgl(self):
        return dgl.unbatch(self._batch.dgl())[self._index]

    def get_type(self):
        return self._batch._graph_type

    def __repr__(self):
        return GraphBase.__repr__(self)
        
    @staticmethod
    def get_batch_type():
        return GraphBatch
        